/*
 * http.c
 *
 * Copyright (C) 2001-2002 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>

#include "openft.h"

#include "nb.h"
#include "http.h"
#include "mime.h"
#include "html.h"

#include "parse.h"

/*****************************************************************************/

/*
 * Object hierarchy:
 *
 *  + HTTPTransfer      (OpenFT)
 *  |--- Chunk          (giFT)
 *  `------ Transfer    (giFT)
 *
 * This hierarchy sucks.  TODO.
 */

/*****************************************************************************/

#define HTTP_USER_AGENT PACKAGE " " VERSION

static HashTable *active_chunks  = NULL;

/* log_format_time */
static const char *month_tab =
	"Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";

/*****************************************************************************/

/* i didnt write this, dont blame me */
static char *log_format_time (time_t log_time)
{
	static char  buf[30];
	char        *p;
	struct tm   *t;
	unsigned int a;
	int          time_offset = 0;

	t = gmtime (&log_time);

	p = buf + 29;

	*p-- = '\0';
	*p-- = ' ';
	*p-- = ']';
	a = abs(time_offset / 60);
	*p-- = '0' + a % 10;
	a /= 10;
	*p-- = '0' + a % 6;
	a /= 6;
	*p-- = '0' + a % 10;
	*p-- = '0' + a / 10;
	*p-- = (time_offset > 0) ? '-' : '+';
	*p-- = ' ';

	a = t->tm_sec;
	*p-- = '0' + a % 10;
	*p-- = '0' + a / 10;
	*p-- = ':';
	a = t->tm_min;
	*p-- = '0' + a % 10;
	*p-- = '0' + a / 10;
	*p-- = ':';
	a = t->tm_hour;
	*p-- = '0' + a % 10;
	*p-- = '0' + a / 10;
	*p-- = ':';
	a = 1900 + t->tm_year;
	while (a)
	{
		*p-- = '0' + a % 10;
		a /= 10;
	}
	/* p points to an unused spot */
	*p-- = '/';
	p -= 2;
	memcpy(p--, month_tab + 4 * (t->tm_mon), 3);
	*p-- = '/';
	a = t->tm_mday;
	*p-- = '0' + a % 10;
	*p-- = '0' + a / 10;
	*p = '[';

	/* should be same as returning buf */
	return p;
}

/* apache-style access.log */
static void http_log (HTTPTransfer *transfer, int status_code,
					  int content_length)
{
	FILE *f;
	char *referer;
	char *user_agent;

	if (!transfer)
		return;

	if (!strncmp (transfer->filename, DATA_DIR, strlen (DATA_DIR)))
		return;

	referer    = dataset_lookup (transfer->header, "Referer");
	user_agent = dataset_lookup (transfer->header, "User-Agent");

	if (!referer)
		referer = "-";

	if (!user_agent)
		user_agent = "-";

	if (!(f = fopen (gift_conf_path ("access.log"), "a")))
	{
		GIFT_ERROR (("failed to open ~/.giFT/access.log"));
		return;
	}

	fprintf (f, "%s - - %s\"%s %s HTTP/1.0\" %i %i \"%s\" \"%s\" \n",
			 net_ip_str (transfer->ip), log_format_time (time (NULL)),
			 (char *) dataset_lookup (transfer->header, "rtype"),
			 transfer->filename, status_code, content_length, referer,
			 user_agent);

	fclose (f);
}

/*****************************************************************************/

static void http_header_free (Dataset *dataset)
{
	hash_table_destroy_free (dataset);
}

/*****************************************************************************/

static void http_transfer_set_file (HTTPTransfer *transfer, char *filename)
{
	if (transfer->filename || transfer->encoded)
	{
		free (transfer->filename);
		free (transfer->encoded);
	}

	transfer->filename = STRDUP (filename);
	transfer->encoded  = url_encode (filename);
}

/*****************************************************************************/

HTTPTransfer *http_transfer_new (HTTPTransferCB cb, Protocol *p,
								 unsigned long ip, unsigned short port,
								 char *filename, Chunk *chunk)
{
	HTTPTransfer *transfer;

	transfer = malloc (sizeof (HTTPTransfer));
	memset (transfer, 0, sizeof (HTTPTransfer));

	transfer->chunk    = chunk;

	transfer->callback = cb;
	transfer->protocol = p;

	transfer->ip       = ip;
	transfer->port     = port;

	http_transfer_set_file (transfer, filename);

	if (chunk)
	{
		transfer->range_start = chunk->start + chunk->transmit;
		transfer->range_stop  = chunk->stop;
	}
	else
	{
		transfer->range_start = 0;
		transfer->range_stop  = 0;
	}

	/* assume that this is not a file transfer until later proven to be */
	transfer->file_transfer = FALSE;

	if (!active_chunks)
		active_chunks = hash_table_new ();

	return transfer;
}

void http_transfer_free (HTTPTransfer *transfer)
{
	/* this chunk is no longer active */
	hash_table_remove (active_chunks, (unsigned long) transfer->chunk);

	free (transfer->filename);
	free (transfer->encoded);

	http_header_free (transfer->header);

	/* for uploads, close the file handle */
	if (transfer->f)
		fclose (transfer->f);

	free (transfer);
}

/*****************************************************************************/

/* parse a string resembling "Key1: Value\r\nKeyn: Value\r\n" into a dataset */
static void http_header_parse (Dataset *dataset, char *http_request)
{
	char *token;

	while ((token = string_sep_set (&http_request, "\r\n")))
	{
		char *key;

		trim_whitespace (token);

		key = string_sep (&token, ": ");
		if (!key || !token)
			continue;

		dataset_insert (dataset, key, STRDUP (token));
	}
}

/* parse the HTTP request off of the header for http_header_parse */
static Dataset *http_request_parse (char *http_request)
{
	Dataset *dataset = NULL;
	char    *request;
	char    *rtype;

	if (!http_request)
		return NULL;

	/* data is expected to be in the format:
	 *
	 *   {GET|PUT|PUSH|HEAD} /blabla[ crap]\r\n
	 */

	rtype   = string_sep     (&http_request, " ");     /* GET|PUT|HEAD */
	request = string_sep_set (&http_request, " \r\n"); /* /blabla */

	if (!request)
		return NULL;

	/* TODO - verify request */

	request = url_decode (request);

	dataset_insert (dataset, "rtype",    STRDUP (rtype));
	dataset_insert (dataset, "request",  request);

	http_header_parse (dataset, http_request);

	return dataset;
}

/* parse the HTTP reply from the header */
static Dataset *http_response_parse (char *http_response)
{
	Dataset *dataset = NULL;
	char    *response;

	response = string_sep_set (&http_response, " \r\n");

	if (!response)
		return NULL;

	dataset_insert (dataset, "response", STRDUP (response));

	http_header_parse (dataset, response);

	return dataset;
}

/*****************************************************************************/

/* parse ?key=value&keyn=value&... into a dataset */
static Dataset *http_query_parse (char *request)
{
	Dataset *dataset = NULL;
	char *token;

	while ((token = string_sep (&request, "&")))
	{
		char *key;

		key = string_sep (&token, "=");
		if (!key || !token)
			continue;

		dataset_insert (dataset, key, STRDUP (token));
	}

	return dataset;
}

/* handle cgi queries */
static char *http_handle_cgiquery (Connection *c, char *request)
{
	unsigned short  listen_port;
	unsigned long   request_host;
	char           *request_file;
	Dataset        *dataset;

	/* parse it -- this should be another func */
	dataset = http_query_parse (request);

	/* only thing we currently use this code for is firewalled uploading */
	listen_port  = ATOI      (dataset_lookup (dataset, "listen_port"));
	request_host = inet_addr (dataset_lookup (dataset, "request_host"));
	request_file = dataset_lookup (dataset, "request_file");

	if (request_host && request_file)
	{
		ft_node_send_indirect (request_host, FT_PUSH_REQUEST, "%lu%hu%s",
							   ntohl (inet_addr (net_peer_ip (c->fd))),
							   listen_port, url_decode (request_file));
	}

	http_header_free (dataset);

	return html_page_redirect (gift_conf_path ("OpenFT/redirect.html"));
}

/*****************************************************************************/

static int http_connection_close (Protocol *p, Connection *c)
{
	HTTPTransfer *transfer;

	/* trace_sock (c->fd, NULL); */

	/* cleanup with the event loop */
	input_remove (c);
	net_close (c->fd);

	transfer = c->data;

	/* notify the transfer function */
	if (transfer->callback)
		(*transfer->callback) (c, transfer->chunk, NULL, 0);

	/* free any data associated */
	http_transfer_free (transfer);

	connection_destroy (c);

	return FALSE;
}

/*****************************************************************************/

/* callback to read from the desired file and write to the requesting
 * client */
static int http_connection_deliver (Protocol *p, Connection *c)
{
	HTTPTransfer *transfer = c->data;
	char       buf[RW_BUFFER + 1];
	size_t     send_len;
	int        n;

	send_len  = fread (buf, sizeof (char), sizeof (buf) - 1, transfer->f);

	/* If an error occurs, or the end-of-file is reached, the return value
	 * is a short item count (or zero).  fread (3) ??? */
	if (send_len > sizeof (buf))
		send_len = 0;

	if (send_len > (transfer->range_stop - transfer->transmit))
		send_len = transfer->range_stop - transfer->transmit;

	if (send_len)
		n = send (c->fd, buf, send_len, 0);

	if (send_len == 0 || n <= 0)
	{
		http_connection_close (p, c);

		return FALSE;
	}

	if (!transfer->chunk)
	{
		char *file;

		if ((file = strrchr (transfer->filename, '/')))
			file++;

		transfer->chunk = malloc (sizeof (Chunk));
		memset (transfer->chunk, 0, sizeof (Chunk));

		/* this is a stupid design.  HTTPTransfer -> Chunk -> Transfer doesnt
		 * make as much sense as Chunk -> Transfer -> HTTPTransfer or
		 * something...sigh...TODO */
		transfer->chunk->transfer =
		    upload_new (p, net_peer_ip (c->fd), file, transfer->filename,
		                transfer->range_stop - transfer->range_start);
	}

	transfer->transmit += n;
	(*transfer->callback) (c, transfer->chunk, buf, send_len);

	return TRUE;
}

/*****************************************************************************/

/* send the appropriate HTTP reply header */
static int http_handle_header (Connection *c, HTTPTransfer *transfer,
							   int nodepage)
{
	char         *http, *range;
	char         *content_type;
	int	          http_status_code = 200;
	unsigned long content_length = 0;

	/* check whether the file is among those shared */
	if (!nodepage && !share_local_find (transfer->filename))
		http_status_code = 404;
	else
	{
		transfer->f = mime_open (transfer->filename, "r",
								 &content_type, &content_length);

		/* if we can't open file its a 404 */
		if (!transfer->f)
			http_status_code = 404;
	}

	/* log this transfer */
	http_log (transfer, http_status_code, content_length);

	if (http_status_code == 404)
	{
		net_send (c->fd, "HTTP/1.0 404 Not Found\r\n\r\n");
		return FALSE;
	}

	/* ok, actually construct the reply now ... this sucks */
	http = malloc (sizeof (char) * 1024);

	range = dataset_lookup (transfer->header, "Range");

	if (range)
	{
		unsigned long start = 0, stop = 0;

		sscanf (range, "bytes=%lu-%lu", &start, &stop);

		if (!stop)
			stop = content_length;

		sprintf (http,
		         "HTTP/1.0 206 Partial Content\r\n"
		         "Content-Range: bytes %lu-%lu/%lu\r\n"
		         "Content-Length: %lu\r\n"
		         "Content-Type: %s\r\n\r\n",
				 start, stop, content_length, stop - start, content_type);

		/* seek to the range start */
		fseek (transfer->f, start, SEEK_SET);

		transfer->range_start = start;
		transfer->range_stop  = stop;

		content_length = stop - start;
	}
	else
	{
		sprintf (http,
		         "HTTP/1.1 200 OK\r\n"
		         "Content-Length: %lu\r\n"
		         "Content-Type: %s\r\n\r\n",
				 content_length, content_type);

		transfer->range_start = 0;
		transfer->range_stop  = content_length;
	}

	send (c->fd, http, strlen (http), 0);
	free (http);

	return TRUE;
}

/*****************************************************************************/
/* TODO clean this mess up */

static int http_read_body (Protocol *p, Connection *c)
{
	HTTPTransfer *transfer = c->data;
	char buf[RW_BUFFER];
	int n;

	if ((n = recv (c->fd, buf, sizeof (buf) - 1, 0)) <= 0)
	{
		trace_sock (c->fd, "remotely closed");

		http_connection_close (p, c);
		return FALSE;
	}

	/* fwrite (buf, sizeof (char), n, transfer->f); */

	/* i should rename this var ;) */
	transfer->transmit += n;
	(*transfer->callback) (c, transfer->chunk, buf, n);

	return TRUE;
}

/* receive the HTTP reply from the socket */
static int http_read_response (Protocol *p, Connection *c)
{
	HTTPTransfer *transfer = c->data;
	NBRead  *nb;
	int      n;

	trace_sock (c->fd, NULL);

	nb = nb_active (c->fd);

	n = nb_read (nb, c->fd, 0, "\r\n\r\n");
	if (n <= 0)
	{
		nb_finish (nb);
		http_connection_close (p, c);
		return FALSE;
	}

	if (!nb->term)
		return FALSE;

	transfer->header = http_response_parse (nb->data);

	nb_finish (nb);

	if (!transfer->header)
	{
		GIFT_WARN (("Invalid HTTP header request from %s",
		            net_peer_ip (c->fd)));
		http_connection_close (p, c);
		return FALSE;
	}

	input_remove (c);
	input_add (p, c, INPUT_READ, (InputCallback) http_read_body);

	return TRUE;
}

/*****************************************************************************/

/* handles incoming request processing */
static int http_connection (Protocol *p, Connection *c)
{
	HTTPTransfer *transfer = c->data;
	char     buf[RW_BUFFER];
	char    *nodepage = NULL;
	char    *rtype;
	int      n;

	/* clear the read queue */
	if ((n = recv (c->fd, buf, sizeof (buf) - 1, 0)) <= 0)
	{
		http_connection_close (p, c);
		return FALSE;
	}

	buf[n] = 0;

	/* parse */
	if (!(transfer->header = http_request_parse (buf)))
	{
		GIFT_WARN (("Invalid HTTP header request from %s",
		            net_peer_ip (c->fd)));
		http_connection_close (p, c);
		return FALSE;
	}

	/* associate the file data with the request */
	http_transfer_set_file (transfer,
							dataset_lookup (transfer->header, "request"));

#if 0
	GIFT_DEBUG (("HTTP request from %s for %s", net_ip_str (transfer->ip),
	             transfer->filename));
#endif

	/* process the request type */
	rtype = dataset_lookup (transfer->header, "rtype");

	/* ok, this is a non-standard OpenFT HTTP session, send the ack then come
	 * back to this connection and read the response */
	if (!strcmp (rtype, "PUSH"))
	{
		trace_sock (c->fd, "push");

		http_header_free (transfer->header);

		GIFT_DEBUG (("directional change: push -> ft_download"));
		transfer->callback = ft_download;

		net_send (c->fd, "HTTP/1.1 200 OK\r\n\r\n");

		input_remove (c);
		input_add (p, c, INPUT_READ, (InputCallback) http_read_response);

		return TRUE;
	}

	/* no file was requested, send the nodepage */
	if (!transfer->filename[1])
	{
		/* nodepage_mtime is set everytime we write out this file, so if it
		 * unequal to the actual mtime of the file, update */
		nodepage = html_page_index (gift_conf_path ("OpenFT/nodepage.html"));
		http_transfer_set_file (transfer, nodepage);
	}
	/* nodepage images */
	else if (!strncmp (transfer->filename, "/OpenFT/", 8))
	{
		char *true_path;

		true_path = malloc (strlen (transfer->filename) + strlen (DATA_DIR) + 1);
		sprintf (true_path, "%s%s", DATA_DIR, transfer->filename);

		http_transfer_set_file (transfer, true_path);

		free (true_path);
	}
	/* CGI query */
	else if (transfer->filename[1] == '?')
	{
		nodepage = http_handle_cgiquery (c, transfer->filename + 2);
		http_transfer_set_file (transfer, nodepage);
	}
	/* file download */
	else
	{
		transfer->file_transfer = TRUE;

		/* make sure we can share the file they requested */
	}

	/* respond */
	if (!http_handle_header (c, transfer, (nodepage ? TRUE : FALSE)))
		http_transfer_set_file (transfer, NULL);

	/* no file was meant to be written, die */
	if (!transfer->filename)
	{
		http_connection_close (p, c);
		return FALSE;
	}

	/* handle the push'd data */
	input_remove (c);
	input_add (p, c, INPUT_WRITE, (InputCallback) http_connection_deliver);

	return TRUE;
}

/*****************************************************************************/

/* incoming connections callback */
void http_handle_incoming (Protocol *p, Connection *c)
{
	Connection *new_c;

	new_c = connection_new (p);
	new_c->fd = net_accept (c->fd);

	/* we dont know much about this connection, but we can fill all that in
	 * as we go */
	new_c->data = http_transfer_new (ft_upload, p,
									 inet_addr (net_peer_ip (new_c->fd)), 0,
									 NULL, NULL);
	input_add (p, new_c, INPUT_READ, (InputCallback) http_connection);
}

/*****************************************************************************/

/* make an HTTP request */
static int http_send_request (int fd, char *rtype, char *request, char *extra)
{
	char  *req;
	size_t req_len = 0;
	int    ret;

	if (!rtype || !request)
		return -1;

	/* just to be safe */
	req = malloc (512 + (extra ? strlen (extra) : 0) + strlen (request));

	/* write the default header */
	req_len += sprintf (req, "%s %s%s HTTP/1.1\r\nUser-Agent: %s\r\n",
						rtype, ((*request != '/') ? "/" : ""), request,
						HTTP_USER_AGENT);

	/* append any arbitrary user data */
	if (extra)
		req_len += sprintf (req + req_len, "%s\r\n", extra);

	/* extra \r\n */
	req_len += sprintf (req + req_len, "\r\n");

	ret = send (fd, req, req_len, 0);

	free (req);

	return ret;
}

/*****************************************************************************/

static int http_push_ack (Protocol *p, Connection *c)
{
	HTTPTransfer *transfer = c->data;
	char     buf[RW_BUFFER];
	int      n;

	trace_sock (c->fd, NULL);

	if ((n = recv (c->fd, buf, sizeof (buf) - 1, 0)) <= 0)
	{
		http_connection_close (p, c);
		return FALSE;
	}

	buf[n] = 0;

	if (!(transfer->header = http_response_parse (buf)))
	{
		GIFT_WARN (("Invalid HTTP header request from %s",
		            net_peer_ip (c->fd)));
		http_connection_close (p, c);
		return FALSE;
	}

	/* TODO - evaluate response */

	/* write the header */
	if (!http_handle_header (c, transfer, FALSE))
	{
		http_connection_close (p, c);
		return FALSE;
	}

	input_remove (c);
	input_add (p, c, INPUT_WRITE, (InputCallback) http_connection_deliver);

	return TRUE;
}

static int http_push_verify_source (Protocol *p, Connection *c)
{
	HTTPTransfer *transfer = c->data;

	trace_sock (c->fd, NULL);

	if (net_sock_error (c->fd))
	{
		http_connection_close (p, c);
		return FALSE;
	}

	http_send_request (c->fd, "PUSH", transfer->encoded, NULL);

	input_remove (c);
	input_add (p, c, INPUT_READ, (InputCallback) http_push_ack);

	return TRUE;
}

/* connect outgoing and "push" the requested file */
int http_push_file (HTTPTransfer *transfer)
{
	Connection *c;

	trace ();

	transfer->file_transfer = TRUE;

	if (!share_local_find (transfer->filename))
	{
		http_transfer_free (transfer);
		return FALSE;
	}

	printf ("push: connecting to %s:%hu\n",
			net_ip_str (transfer->ip), transfer->port);

	c = connection_new (transfer->protocol);
	c->fd = net_connect (net_ip_str (transfer->ip), transfer->port);

	c->data = transfer;
	input_add (transfer->protocol, c, INPUT_WRITE,
			   (InputCallback) http_push_verify_source);

	hash_table_insert (active_chunks, (unsigned long) transfer->chunk, c);

	return TRUE;
}

/*****************************************************************************/

static int http_pull_verify_source (Protocol *p, Connection *c)
{
	HTTPTransfer *transfer = c->data;
	char *range = NULL;

	trace_sock (c->fd, NULL);

	if (net_sock_error (c->fd))
	{
		http_connection_close (p, c);
		return FALSE;
	}

	if (transfer->range_stop)
	{
		range = malloc (128);
		sprintf (range, "Range: bytes=%lu-%lu\r\n",
				 transfer->range_start, transfer->range_stop);
	}

	http_send_request (c->fd, "GET", transfer->encoded, range);

	free (range);

	input_remove (c);
	input_add (p, c, INPUT_READ, (InputCallback) http_read_response);

	return TRUE;
}

int http_pull_file (HTTPTransfer *transfer)
{
	Connection *c;

	trace ();

	transfer->file_transfer = TRUE;

	printf ("%s: connecting to %s:%hu\n", __PRETTY_FUNCTION__,
			net_ip_str (transfer->ip), transfer->port);

	c = connection_new (transfer->protocol);
	c->fd = net_connect (net_ip_str (transfer->ip), transfer->port);

	c->data = transfer;
	input_add (transfer->protocol, c, INPUT_WRITE,
			   (InputCallback) http_pull_verify_source);

	hash_table_insert (active_chunks, (unsigned long) transfer->chunk, c);

	return TRUE;
}

/*****************************************************************************/

void http_cancel (Chunk *chunk)
{
	Connection   *http_c;
	HTTPTransfer *http_xfer;

	trace ();

	/* TODO - use a unique id */
	if (!(http_c = hash_table_lookup (active_chunks, (unsigned long) chunk)))
		return;

	http_xfer = http_c->data;

	printf ("destroying %p\n", http_xfer);

	/* callback should NOT be made, we were instructed to remove this data */
	http_xfer->callback = NULL;
	http_connection_close (http_xfer->protocol, http_c);
}
