/*
 * netorg.c - core OpenFT network organization
 *
 * Copyright (C) 2001-2002 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "openft.h"

#include "netorg.h"

/*****************************************************************************/

static List *connections = NULL;
static List *iterator = NULL;

/*****************************************************************************/

void conn_add (Connection *c)
{
	if (!c)
		return;

	connections = list_append (connections, c);

	if (!iterator)
		iterator = connections;
}

void conn_remove (Connection *c)
{
	if (!c)
		return;

	if (iterator && iterator->data == c)
		iterator = connections;

	connections = list_remove (connections, c);

	/* if connections length is 1, c is the head and tail, therefore confusing
	 * the logic above.  force it correctly here */
	if (!connections)
		iterator = NULL;
}

/*****************************************************************************/

static Connection *node_reconnect (Connection *c, Node *node, int *processed)
{
#if 0
	printf ("*** reconnecting %s:%hu\n", net_ip_str (node->ip), node->port);
#endif

	ft_reconnect (c->protocol, c);

	if (processed)
		(*processed)++;

	return NULL;
}

void conn_maintain (Protocol *p)
{
	/* constantly retry a small amount of user nodes to robust the list */
	conn_foreach ((ConnForeachFunc) node_reconnect, NULL,
	              NODE_NONE, NODE_DISCONNECTED, 10);

	/* we aren't satisfied */
	if (conn_length (NODE_SEARCH, NODE_CONNECTED) < 15)
	{
		int processed;

		conn_foreach ((ConnForeachFunc) node_reconnect, &processed,
		              NODE_SEARCH, NODE_DISCONNECTED, 15);

		/* we still aren't satisfied */
		if (processed < 15)
		{
			/* try 10 more users */
			conn_foreach ((ConnForeachFunc) node_reconnect, NULL,
			              NODE_NONE, NODE_DISCONNECTED, 10);
		}
	}
}

/*****************************************************************************/

/* connection iterator */
Connection *conn_foreach (ConnForeachFunc func, void *user_data,
                          unsigned short klass, unsigned short state,
                          unsigned short iter)
{
	Connection *ret = NULL;
	Connection *conn;
	Node       *node;
	List       *ptr;
	int         looped = FALSE;

	assert (func != NULL);

	/* if iteration was supplied, use iterator, otherwise ignore it */
	ptr = (iter) ? iterator : connections;

	/* this is cute */
	for (;; ptr = list_next (ptr))
	{
		/* if iter, start over from the beginning of the list */
		if (iter && !ptr && !looped)
		{
			ptr = connections;
			looped = TRUE;
		}

		if (!ptr)
			break;

		conn = ptr->data;
		node = NODE (conn);

#if 0
		printf ("*** evaluating %s:%hu (%hu, %hu, %i)\n",
				net_ip_str (node->ip), node->port,
				klass, state, iter);
#endif

		/* filter out the conditions supplied */
		if (klass && !(node->class & klass))
			continue;

		if (state != (unsigned short) -1 && node->state != state)
			continue;

		/* callback */
		ret = (*func) (conn, node, user_data);

#if 0
		printf ("*** %p (%p)\n", func, ret);
#endif

		/* iter may have been passed as 0, if that is the case, ignore its
		 * behavior completely */
		if (iter)
		{
			iterator = ptr->next;
			iter--;

			if (!iter)
				break;
		}

		if (ret)
			break;
	}

	return ret;
}

/*****************************************************************************/

static Connection *conn_counter (Connection *c, Node *node, int *length)
{
	(*length)++;
	return NULL;
}

int conn_length (unsigned short klass, unsigned short state)
{
	int length = 0;

	conn_foreach ((ConnForeachFunc) conn_counter, &length, klass, state, 0);

	return length;
}

/*****************************************************************************/

/* prevent multiple connections to the same host */
int conn_auth (char *ip, unsigned short port)
{
	List       *ptr, *next;
	Connection *loop_conn;
	char       *comp_ip;
	int         ret = 1;

	/* obviously this is invalid. */
	if (!strcmp (ip, "0.0.0.0") || port == 0)
		return 0;

	/* they probably used net_ip_str, which is a no-no here. */
	comp_ip = strdup (ip);

	for (ptr = connections; ptr; ptr = next)
	{
		next = ptr->next;

		if (!(loop_conn = ptr->data))
			continue;

		if (!strcmp (net_ip_str (NODE (loop_conn)->ip), comp_ip))
		{
			/* we found a match, but this node is dead, so remove it and keep
			 * searching */
			if (NODE (loop_conn)->state == NODE_DISCONNECTED ||
				NODE (loop_conn)->class == NODE_NONE)
			{
#if 0
				GIFT_DEBUG (("removing stale connection %s:%hu", comp_ip,
							 port));
#endif

				/* disable this node (possibly remove) */
				node_remove_connection (NULL, loop_conn);

				/* the function above doesnt require the loop_conn be
				 * removed from the list, so make sure that happens */
				if (list_find (connections, loop_conn))
				{
					conn_remove (loop_conn);

					free (loop_conn->data);
					connection_destroy (loop_conn);
				}

				continue;
			}

#if 0
			/* it is already a valid connection, get out of here with an
			 * error */
			GIFT_DEBUG (("refusing authorization for connection to %s:%hu",
						 ip, port));
#endif

			ret = 0;
			break;
		}
	}

#if 0
	if (ret)
		GIFT_DEBUG (("accepting authorization for connection to %s:%hu",
					 ip, port));
#endif

	free (comp_ip);
	return ret;
}
