/*
 * node.c
 *
 * Copyright (C) 2001-2002 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <stdarg.h> /* va lists */
#include <time.h>   /* time ()  */

#include "openft.h"

#include "netorg.h"

#include "queue.h"
#include "nb.h"

#include "http.h"
#include "html.h"

/**************************************************************************/

/**/extern Connection *ft_self;

/**************************************************************************/

static int ft_node_send_constructed (Connection *c, unsigned short ft_command,
									 char *data, int len)
{
	FTPacket *packet;

	packet = ft_packet_new (ft_command, data, len);

	free (data);

	if (!packet)
		return FALSE;

	/* we aren't going to manipulate this buffer, so send it */
	ft_packet_send (c, packet);
	ft_packet_free (packet);

	return TRUE;
}

/*
 * send a command to the supplied node connection
 *
 * request may be NULL for a 0 length packet ( + 4 bytes for length+command)
 */
int ft_node_send (Connection *c, unsigned short ft_command, char *fmt, ...)
{
	va_list args;
	char *data = NULL;
	int   len  = 0;

	if (fmt)
	{
		va_start (args, fmt);
		data = ft_packet_data (fmt, args, &len);
		va_end (args);
	}

	return ft_node_send_constructed (c, ft_command, data, len);
}

static Connection *locate_node (Connection *c, Node *node,
								unsigned long deliver_to)
{
	if (node->ip == deliver_to)
		return c;

	return NULL;
}

/*
 * send a message to the supplied child
 */
int ft_node_send_indirect (unsigned long deliver_to, unsigned short ft_command,
						   char *fmt, ...)
{
	Connection *c;
	va_list args;
	char *data = NULL;
	int   len  = 0;

	/* locate the connection to send this to */
	c = conn_foreach ((ConnForeachFunc) locate_node, (void *) deliver_to,
	                  NODE_USER, NODE_CONNECTED, 0);

	if (!c)
		return -1;

	if (fmt)
	{
		va_start (args, fmt);
		data = ft_packet_data (fmt, args, &len);
		va_end (args);
	}

	return ft_node_send_constructed (c, ft_command, data, len);
}

/*
 * main node communication loop
 *
 * handle reading of packets from a node connection (at this point
 * indiscriminant as to whether it is a search or user node)
 */
void ft_node_connection (Protocol *p, Connection *c)
{
	NBRead *nb;
	unsigned short len;
	int n;

	/* we are connected if we got here ... */
	node_state_set (c, NODE_CONNECTED);
	NODE (c)->vitality = time (NULL);

	/* grab a data buffer */
	nb = nb_active (c->fd);

	/* ok this nasty little hack here is executed so that we can maintain
	 * a single command buffer for len/command and the linear data
	 * stream to follow */
	len = nb->flag + FT_HEADER_LEN;

	n = nb_read (nb, c->fd, len, NULL);
	if (n <= 0)
	{
		/* nevermind, we are NOT connected :) */
		nb_finish (nb);
		node_remove_connection (p, c);
		return;
	}

	/* we successfully read a portion of the current packet, but not all of
	 * it ... come back when we have more data */
	if (!nb->term)
		return;

	/* packet length */
	len = ntohs (*((ft_uint16 *)nb->data));

	if (nb->flag || len == 0) /* we read a complete packet */
	{
		FTPacket *packet;
		int       ret;

		/* this code is totally bastardized. */
		packet = ft_packet_new (ntohs (*((ft_uint16 *)(nb->data + 2))),
		                        nb->data + FT_HEADER_LEN, len);

		ret = protocol_handle_command (p, c, packet);

		ft_packet_free (packet);

		/* we're done with this scratch buffer, get rid of it */
		nb_finish (nb);

		/* TODO - should we do this? */
		if (!ret)
		{
			node_remove_connection (p, c);
			return;
		}
	}
	else if (!nb->flag) /* we read a header block with non-zero length waiting
						 * on the socket */
	{
		nb->term = 0;
		nb->flag = len;
	}
}

/*****************************************************************************/

unsigned short node_highest_class (Node *node)
{
	unsigned short high_class;

	if (node->class & NODE_INDEX)
		high_class = NODE_INDEX;
	else if (node->class & NODE_SEARCH)
		high_class = NODE_SEARCH;
	else if (node->class & NODE_USER)
		high_class = NODE_USER;
	else
		high_class = NODE_NONE;

	return high_class;
}

/*****************************************************************************/

Connection *node_add_connection (Protocol *p, int fd)
{
	Connection *c = connection_new (p);
	Node *nc;
	struct sockaddr_in saddr;
	int len;

	conn_add (c);

	/* fill the connection data */
	c->data = nc = malloc (sizeof (Node));
	memset (nc, 0, sizeof (Node));

	/* fill the peer information for badly misbehaving connections or
	 * accept()'d connections */
	if (fd >= 0 && !nc->ip)
	{
		len = sizeof (saddr);
		if (getpeername (fd, (struct sockaddr *)&saddr, &len) == 0)
			nc->ip = saddr.sin_addr.s_addr;
	}

#if 0
	GIFT_DEBUG (("added connection %i from %s:%hu", fd,
	             net_ip_str (nc->ip), nc->port));
#endif

	/* set fd */
	c->fd = fd;
	nc->fd = fd;

	nc->incoming = FALSE;

	return c;
}

void node_remove_connection (Protocol *p, Connection *c)
{
	/* remove from select and free */
	input_remove (c);

	net_close (c->fd);

#if 0
	GIFT_DEBUG (("removed connection %i from %s:%hu", c->fd,
				 net_ip_str (NODE (c)->ip), NODE (c)->port));
#endif

	/* if this node was a child, get rid of all it's shares */
	if (NODE (c)->class & NODE_CHILD)
	{
		share_remove_by_host (net_ip_str (NODE (c)->ip));
		node_class_remove (c, NODE_CHILD);
	}

	node_state_set (c, NODE_DISCONNECTED);
	NODE (c)->fd = c->fd = -1;

	/* get rid of any queued writes */
	queue_free (c);

#if 0
	/* this function is a bit tricky on the back-end... we will keep the
	 * connection in this list until it grows beyond the default, and then
	 * we will actually start removing it */
	if (list_length (ft_connections) > 1000)
	{
		/* if it is a child, remove it from the connections list */
		ft_connections = list_remove (ft_connections, c);
		free (c->data);
		connection_destroy (c);
	}
#endif
}

/*****************************************************************************/

static Connection *node_maintain_link (Connection *c, Node *node, FILE *f)
{
	fprintf (f, "%lu %s:%hu %hu\n", node->vitality, net_ip_str (node->ip),
			 node->port, node->http_port);

	return NULL;
}

static Connection *node_stats (Connection *c, Node *node, void *data)
{
	ft_node_send (c, FT_STATS_REQUEST, NULL);

	return NULL;
}

int node_maintain_links (Protocol *p)
{
	FILE *f;

	/* gather stats information */
	conn_foreach ((ConnForeachFunc) node_stats, NULL,
	              NODE_SEARCH, NODE_CONNECTED, 0);

	conn_maintain (p);

	f = fopen (gift_conf_path ("OpenFT/nodes"), "w");

	if (!f)
		perror ("fopen");
	else
	{
		conn_foreach ((ConnForeachFunc) node_maintain_link, f,
					  NODE_NONE, -1, 0);

		fclose (f);
	}

	return TRUE;
}

/*****************************************************************************/

static void node_class_change (Connection *c, unsigned short change)
{
	/* ignore local class changes...expect that they are handled on an
	 * individual level */
	if (c == ft_self)
		return;

	if (!change)
		return;

	printf ("%s:%hu changed class to NODE_%s\n",
			net_ip_str (NODE (c)->ip), NODE (c)->port,
			node_class_str (NODE (c)));

	/* if the node has just changed to a search node */
	if (change & NODE_SEARCH)
		share_local_submit (c->protocol, c);

	/* children should increment total users */
	if (change & NODE_CHILD)
		NODE (ft_self)->users++;

	html_update_nodepage ();
}

void node_class_set (Connection *c, unsigned short class)
{
	unsigned short orig;

	orig = NODE (c)->class;
	NODE (c)->class = class;
	node_class_change (c, (class & ~orig));
}

void node_class_add (Connection *c, unsigned short class)
{
	unsigned short orig;

	orig = NODE (c)->class;
	NODE (c)->class |= class;
	node_class_change (c, (class & ~orig));
}

void node_class_remove (Connection *c, unsigned short class)
{
	if (class & NODE_CHILD)
		NODE (ft_self)->users--;

	NODE (c)->class &= ~class;
}

/*****************************************************************************/

void node_state_set (Connection *c, unsigned short state)
{
	char *state_str;
	int   change = 0;

	if (NODE (c)->state != state)
		change = 1;

	NODE (c)->state = state;

	/* dont print useless info */
	if (!change || !NODE (c)->class || NODE (c)->vitality + 300 < time (NULL))
		return;

	switch (state)
	{
	 case NODE_CONNECTED:
		state_str = "CONNECTED";
		break;
	 case NODE_CONNECTING:
		state_str = "CONNECTING";
		break;
	 default:
		state_str = "DISCONNECTED";
		break;
	}

	printf ("%s:%hu changed state to NODE_%s\n",
			net_ip_str (NODE (c)->ip), NODE (c)->port,
			state_str);

	/* if we just got connected, emulate a class change */
	if (state == NODE_CONNECTED)
		node_class_change (c, NODE (c)->class);

	html_update_nodepage ();
}

/*****************************************************************************/

void node_conn_set (Connection *c, unsigned long ip, signed long port,
					signed long http_port)
{
	if (ip)
		NODE (c)->ip = ip;

	if (port >= 0)
	{
		NODE (c)->firewalled = (port == 0);
		NODE (c)->port = port;
	}

	if (http_port > 0)
		NODE (c)->http_port = http_port;

	html_update_nodepage ();
}
