/*
 * packet.c
 *
 * Copyright (C) 2001-2002 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "openft.h"

/**************************************************************************/

#define PACKET_APPEND(str,len,data,data_len) \
        if (len + data_len < 4096) \
        { \
        	memcpy (str + len, data, data_len); \
        	len += data_len; \
		}

/*****************************************************************************/

/*
 * construct a new OpenFT packet
 */
FTPacket *ft_packet_new (ft_uint16 ft_command, char *data, int len)
{
	FTPacket *packet;

	assert (len >= 0);

	/* allocate enough size to include the request with the structure in
	 * a linear fashion */
	packet = malloc (sizeof (FTPacket) + len);

	packet->len     = len;
	packet->command = ft_command;
	packet->offset  = 0;

	/* move the data onto the structures buffer */
	if (data)
	{
		memcpy (packet->data, data, len);
		memset (packet->data + len, 0, sizeof (char));
	}

#if 0
	printf ("packet_new: %hu %hu { ", len, ft_command);

	{
		int i;
		for (i = 0; i < packet->len; i++)
			printf ("0x%02x ", packet->data[i]);
	}

	printf ("}\n");
#endif

	return packet;
}


/*
 * clean up.
 */
void ft_packet_free (FTPacket *packet)
{
	free (packet);
}

/*****************************************************************************/

/*
 * construct the data string for an OpenFT packet
 *
 * printf-like arguments translate to raw binary net types
 *
 * TODO - optimize this
 */
char *ft_packet_data (char *fmt, va_list args, int *len)
{
	char  c;
	char *output  = malloc (4096);
	int   out_len = 0;
	char *str;
	ft_uint32 ul;
	ft_uint16 us;

	assert (fmt);
	assert (*fmt);

	while (*fmt)
	{
		if (*fmt++ == '%')
		{
			/* *fmt will be the possible long opt, c being the modifier */
			c = *fmt++;

			switch (c)
			{
			 case 'h': /* short */
				us = htons ((unsigned short) va_arg (args, long));
				PACKET_APPEND (output, out_len, &us, sizeof (us));
				fmt++;
				break;
			 case 'l': /* long */
				ul = htonl ((unsigned long) va_arg (args, long));
				PACKET_APPEND (output, out_len, &ul, sizeof (ul));
				fmt++;
				break;
			 case 's': /* ascii string */
				str = va_arg (args, char *);
				/* include \0 */
				PACKET_APPEND (output, out_len, str, strlen (str) + 1);
				break;
			}
		}
	}

	if (len)
		*len = out_len;

	/* we now have a string ready for network transmission */
	return output;
}

ft_uint32 ft_packet_get_int (FTPacket *packet, size_t size, int host_order)
{
	char      data_8;
	ft_uint16 data_16;
	ft_uint32 data_32 = 0;
	char     *offs;

	assert (packet);

	/* check overrun */
	if (packet->offset + size > packet->len)
		return 0;

	offs = packet->data + packet->offset;

	switch (size)
	{
	 case 1:
		memcpy (&data_8, offs, size);
		data_32 = (ft_uint32) data_8;
		break;
	 case 2:
		memcpy (&data_16, offs, size);
		data_32 = (ft_uint32) ((host_order ? ntohs (data_16) : data_16));
		break;
	 case 4:
		memcpy (&data_32, offs, size);
		if (host_order)
			data_32 = ntohl (data_32);
		break;
	 default:
		printf ("%s: wtf are you doing?\n", __PRETTY_FUNCTION__);
		return data_32;
	}

	packet->offset += size;

	return data_32;
}

char *ft_packet_get_str (FTPacket *packet)
{
	char *start, *ptr;
	int i;

	assert (packet);

	if (packet->offset == packet->len)
		return NULL;

	/* we have to write our own strlen function so that we can't be DoS'd
	 * attack of the script kiddies can really hurt :) */
	i      = packet->len - packet->offset;
	start  = ptr = packet->data + packet->offset;

	while (*ptr++ && i-- > 0);

	/* we ran out of data before we found the \0, treat this as an invalid
	 * packet */
	if (i <= 0)
	{
		packet->offset = packet->len;  /* no subsequent calls will allow this
		                                * packet anymore */
		return NULL;
	}

	/* length is (ptr - start), including the \0 */
	packet->offset += (ptr - start);

	return start;
}

/*****************************************************************************/

/*
 * sends a constructed packet
 *
 * TODO - buffer this and send when the socket is ready for it
 */
int ft_packet_send (Connection *c, FTPacket *packet)
{
	int ret, len;

	/* nothing to send, piss off */
	if (!packet)
		return FALSE;

	len = packet->len;

	/* we can't blindly send this data yet, it's in host order */
	packet->len = htons (packet->len);
	packet->command = htons (packet->command);

	/* packet->len is the start of the data stream that is expected to be
	 * written out */
	ret = send (c->fd, &packet->len, len + FT_HEADER_LEN, 0);

	/* convert back */
	packet->len = ntohs (packet->len);
	packet->command = ntohs (packet->command);

	return ret;
}
