/*
 * protocol.c
 *
 * Copyright (C) 2001-2002 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "openft.h"

#include "netorg.h"

#include "http.h"

#include "search.h"
#include "parse.h"

/**************************************************************************/

/**/extern Connection *ft_self;

/**************************************************************************/

typedef void (*ProtocolHandler) (Protocol *p, Connection *c, FTPacket *packet);
#define FT_HANDLER(func) static void ft_##func (Protocol *p, Connection *c, FTPacket *packet)

FT_HANDLER (class_request);
FT_HANDLER (class_response);

FT_HANDLER (share_request);
FT_HANDLER (share_response);

FT_HANDLER (search_request);
FT_HANDLER (search_response);

FT_HANDLER (nodelist_request);
FT_HANDLER (nodelist_response);

FT_HANDLER (push_request);
FT_HANDLER (push_response);

FT_HANDLER (nodeinfo_request);
FT_HANDLER (nodeinfo_response);

FT_HANDLER (stats_request);
FT_HANDLER (stats_response);

/**************************************************************************/

struct _handler_table
{
	ft_uint16       command;
	ProtocolHandler func;
}
protocol_handler[] =
{
	{ FT_CLASS_REQUEST,     ft_class_request },
	{ FT_CLASS_RESPONSE,    ft_class_response },

	{ FT_SHARE_REQUEST,     ft_share_request },
	{ FT_SHARE_RESPONSE,    ft_share_response },

	{ FT_SEARCH_REQUEST,    ft_search_request },
	{ FT_SEARCH_RESPONSE,   ft_search_response },

	{ FT_NODELIST_REQUEST,  ft_nodelist_request },
	{ FT_NODELIST_RESPONSE, ft_nodelist_response },

	{ FT_PUSH_REQUEST,      ft_push_request },
	{ FT_PUSH_RESPONSE,     ft_push_response },

	{ FT_NODEINFO_REQUEST,  ft_nodeinfo_request },
	{ FT_NODEINFO_RESPONSE, ft_nodeinfo_response },

	{ FT_STATS_REQUEST,     ft_stats_request },
	{ FT_STATS_RESPONSE,    ft_stats_response },

	{ 0,                    NULL }
};

/**************************************************************************/

int protocol_handle_command (Protocol *p, Connection *c, FTPacket *packet)
{
	struct _handler_table *handler;

	if (!packet)
		return FALSE;

#if 0
	{
		int i;
		printf ("%s: %hu, %hu, { ", __PRETTY_FUNCTION__,
				packet->len, packet->command);

		/* temp -- debugging */
		for (i = 0; i < packet->len; i++)
			printf ("0x%02x ", packet->data[i]);

		printf ("}\n");
	}
#endif

	/* locate the handler */
	for (handler = protocol_handler; handler->func; handler++)
	{
		if (packet->command == handler->command)
		{
			(*handler->func) (p, c, packet);
			return TRUE;
		}
	}

	GIFT_ERROR (("[%s] found no handler for %hu %hu",
				 net_ip_str (NODE (c)->ip), packet->len, packet->command));

	return FALSE;
}

/**************************************************************************/

FT_HANDLER (class_request)
{
	ft_node_send (c, FT_CLASS_RESPONSE, "%hu", NODE (ft_self)->class);
}

FT_HANDLER (class_response)
{
	if (packet->len < 2)
		return;

	node_class_set (c, ft_packet_get_int (packet, 2, TRUE));
}

/**************************************************************************/

FT_HANDLER (share_request)
{
	FileShare *file;
	ft_uint32  size;
	char      *md5;
	char      *filename;

#if 0
	/* if we're a search node and have fewer than 500 (TODO - configurable)
	 * children nodes, accept */
	if (!(NODE (c)->class & NODE_CHILD) &&
        (!(NODE (ft_self)->class & NODE_SEARCH) ||
	     conn_length (NODE_CHILD) >= 500))
			return;
#endif

	/* flushed all files...no longer a child of ours */
	if (packet->len == 0)
	{
		share_remove_by_host (net_ip_str (NODE (c)->ip));
		node_class_remove (c, NODE_CHILD);
		return;
	}

	size     = ft_packet_get_int (packet, 4, TRUE);
	md5      = ft_packet_get_str (packet);
	filename = ft_packet_get_str (packet);

	file = share_new (net_ip_str (NODE (c)->ip), NODE (c)->port,
	                  NODE (c)->http_port, size, md5, filename);

	/* new child */
	node_class_add (c, NODE_CHILD);

	share_add (file);
}

FT_HANDLER (share_response)
{
	trace_sock (c->fd, "invalid packet, sorta");
}

/**************************************************************************/

static void search_request_result (Connection *c, FileShare *file,
								   unsigned long id)
{
	if (!file)
	{
		printf ("finished reply for search %lu\n", id);
		ft_node_send (c, FT_SEARCH_RESPONSE, "%lu", id);
		return;
	}

	/* dont send shares they submitted */
	if (NODE (c)->ip != file->host)
	{
		ft_node_send (c, FT_SEARCH_RESPONSE, "%lu%lu%hu%hu%lu%s%s", id,
		              ntohl (file->host), file->port, file->http_port,
		              file->size, file->md5, file->file);
	}
}

static void search_request_result_free (Connection *c, FileShare *file,
										unsigned long id)
{
	/* share_free (file); */
}

FT_HANDLER (search_request)
{
	List      *ptr;
	ft_uint32  id;
	ft_uint16  type;
	char      *query;

	if (!(NODE (ft_self)->class & NODE_SEARCH))
		return;

	id    = ft_packet_get_int (packet, 4, TRUE);
	type  = ft_packet_get_int (packet, 2, TRUE);
	query = ft_packet_get_str (packet);

	printf ("%s (%s:%hu): [%lu:%i] query = '%s'...", __PRETTY_FUNCTION__,
			net_ip_str (NODE (c)->ip), NODE (c)->port, id, type, query);

	ptr = ft_search (type | SEARCH_LOCAL, query);
	printf ("%i results\n", list_length (ptr));

	queue_add (c,
			   (QueueWriteFunc) search_request_result,
			   (QueueWriteFunc) search_request_result_free,
			   ptr, (void *) id);
}

FT_HANDLER (search_response)
{
	FileShare *file = NULL;
	ft_uint32  id;
	ft_uint32  host;
	ft_uint16  port;
	ft_uint16  http_port;
	ft_uint32  size;
	char      *md5;
	char      *filename;

	id = ft_packet_get_int (packet, 4, TRUE);

	/* packet->len == 4 when an EOF came in */
	if (packet->len > 4)
	{
		host      = ft_packet_get_int (packet, 4, FALSE);
		port      = ft_packet_get_int (packet, 2, TRUE);
		http_port = ft_packet_get_int (packet, 2, TRUE);
		size      = ft_packet_get_int (packet, 4, TRUE);
		md5       = ft_packet_get_str (packet);
		filename  = ft_packet_get_str (packet);

		if (!host)
			host = NODE (c)->ip;

		file = share_new (net_ip_str (host), port, http_port,
		                  size, md5, filename);

		/* host may have lied...if this is the case, lets fix it. */
		if (file->host == 0)
			file->host = NODE (c)->ip;
	}

	search_reply (id, c, file);

	share_free (file);
}

/**************************************************************************/

static Connection *nodelist_output (Connection *c, Node *node,
                                    Connection *data)
{
#if 0
	printf ("nodelist_request: sending out %s:%hu\n",
			net_ip_str (NODE (c)->ip), NODE (c)->port);
#endif

	if (NODE (c)->ip != NODE (data)->ip)
	{
		/* ip is in network order, so we need to swap it so that it
		 * can be swapped again in ft_node_send ... */
		ft_node_send (data, FT_NODELIST_RESPONSE, "%lu%hu",
					  ntohl (NODE (c)->ip), NODE (c)->port);
	}

	return NULL;
}

FT_HANDLER (nodelist_request)
{
	conn_foreach ((ConnForeachFunc) nodelist_output, c,
				  NODE_USER, NODE_CONNECTED, 0);
}

FT_HANDLER (nodelist_response)
{
	ft_uint32 ip;
	ft_uint16 port;

	/* ip should remain in network order */
	ip   = ft_packet_get_int (packet, 4, FALSE);
	port = ft_packet_get_int (packet, 2, TRUE);

	ft_connect (p, net_ip_str (ip), port);
}

/*****************************************************************************/

FT_HANDLER (push_request)
{
	HTTPTransfer  *transfer;
	unsigned long  ip;
	unsigned short port;
	char          *file;

	ip   = ft_packet_get_int (packet, 4, FALSE);
	port = ft_packet_get_int (packet, 2, TRUE);
	file = ft_packet_get_str (packet);

	/* port == 0 in this instance means that we are to send this file back to
	 * the search node contacting us here */
	if (!port)
	{
		ip   = inet_addr (net_ip_str (NODE (c)->ip));
		port = NODE (c)->http_port;
	}

	printf ("received push request for %s:%hu, %s\n",
	        net_ip_str (ip), port, file);

	transfer = http_transfer_new (ft_upload, p, ip, port, file, NULL);
	http_push_file (transfer);
}

FT_HANDLER (push_response)
{
	trace_sock (c->fd, "undefined function");
}

/*****************************************************************************/

FT_HANDLER (nodeinfo_request)
{
	/* this function is for padding...currently only port is actually
	 * transmitted, IP is 0 */
	ft_node_send (c, FT_NODEINFO_RESPONSE, "%lu%hu%hu", 0,
				  NODE (ft_self)->port, NODE (ft_self)->http_port);
}

FT_HANDLER (nodeinfo_response)
{
	unsigned long ip;
	unsigned short port;
	unsigned short http_port;

	ip        = ft_packet_get_int (packet, 4, FALSE);
	port      = ft_packet_get_int (packet, 2, TRUE);
	http_port = ft_packet_get_int (packet, 2, TRUE);

	printf ("-> %lu:%hu (%hu)\n", ip, port, http_port);

	node_conn_set (c, 0, port, http_port);
}

/*****************************************************************************/

FT_HANDLER (stats_request)
{
	ft_node_send (c, FT_STATS_RESPONSE, "%lu%lu%lu",
				  NODE (ft_self)->users, NODE (ft_self)->shares,
				  NODE (ft_self)->megs);
}

FT_HANDLER (stats_response)
{
	unsigned long users, shares, megs;

	users  = ft_packet_get_int (packet, 4, TRUE);
	shares = ft_packet_get_int (packet, 4, TRUE);
	megs   = ft_packet_get_int (packet, 4, TRUE);

	share_stats_set (c, users, shares, megs);
}
