/*
 * share.c
 *
 * Copyright (C) 2001-2002 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

/* TODO - optimize and find a more elegant way to handle local user shares
 * vs search node shares */

#include "openft.h"

#include "queue.h"
#include "search.h"

#include "http.h"
#include "html.h"

#include "parse.h"
#include "file.h"

/*****************************************************************************/

/**/extern Connection *ft_self;
/**/extern Config     *openft_conf;

/* files other nodes are sharing to this one */
List *ft_shares = NULL;

/* files this node is sharing to others */
HashTable *ft_local_shares = NULL;

/*****************************************************************************/

int share_complete (FileShare *file)
{
	int ret = FALSE;

	if (!file)
		return FALSE;

	if (file->file && file->md5 && file->size)
	{
		if (file->file[0] != '/')
			return FALSE;

		if (!file->tokens)
		{
			/* tokenize this query for fast searching */
			file->tokens = search_tokenize (file->file);
		}

		if (file->tokens)
			ret = TRUE;
	}

	return ret;
}

void share_free (FileShare *file)
{
	if (!file)
		return;

	free (file->file);
	free (file->md5);
	free (file->tokens);
	free (file);
}

FileShare *share_new (char *host, unsigned short port,
                      unsigned short http_port, unsigned long size,
                      char *md5, char *filename)
{
	FileShare *file;

	file = malloc (sizeof (FileShare));
	memset (file, 0, sizeof (FileShare));

	file->host      = inet_addr (host);
	file->port      = port;
	file->http_port = http_port;

	file->size = size;
	file->md5  = STRDUP (md5);
	file->file = STRDUP (filename);

	file->tokens = NULL;

	return file;
}

void share_add (FileShare *file)
{
	if (!share_complete (file))
	{
		share_free (file);
		return;
	}

	ft_shares = list_append (ft_shares, file);
	share_stats_add (ft_self, file);
}

void share_remove (FileShare *file)
{
	ft_shares = list_remove (ft_shares, file);
	share_stats_remove (ft_self, file);

	share_free (file);
}

void share_remove_by_host (char *host)
{
	List *ptr = ft_search (SEARCH_HOST, host);

	printf ("removed shares from %s\n", host);

	/* remove all shares found by this user */
	for (; ptr; ptr = ptr->next)
		share_remove ((FileShare *) ptr->data);
}

/*****************************************************************************/
/* local shares */

static int share_local_clear_node (unsigned long key, FileShare *file,
                                   void *udata)
{
	share_free (file);

	return TRUE;
}

void share_local_clear ()
{
	hash_table_foreach_remove (ft_local_shares,
	                           (HashFunc) share_local_clear_node, NULL);
	ft_local_shares = NULL;

	share_stats_reset (ft_self);
}

void share_local_add (FileShare *file)
{
	FileShare *old_file;

	if (!share_complete (file))
		return;

	if (!ft_local_shares)
		ft_local_shares = hash_table_new ();

	old_file = hash_table_lookup (ft_local_shares, hash_string (file->md5));
	if (old_file)
		share_free (old_file);

	hash_table_insert (ft_local_shares, hash_string (file->md5), file);

	/* increment stats */
	share_stats_add (ft_self, file);
}

/*****************************************************************************/

static void path_traverse (char *path)
{

}

void share_local_rebuild ()
{
	char *root0, *root;
	char *path;

	if (!(root = config_get_str (openft_conf, "sharing/root")))
		return;

	root0 = root;

	trace ();

	while ((path = string_sep (&root, " ")))
	{
		if (!file_exists (path) || *path != '/')
		{
			printf ("ignoring '%s'...\n", path);
			continue;
		}

		printf ("rebuilding '%s'...\n", path);

		path_traverse (path);
	}

	free (root0);
}

/*****************************************************************************/

void share_local_import ()
{
	FILE *f;
	char *buf = NULL;
	unsigned long size;
	FileShare *file;

	html_update_nodepage ();
	share_local_clear ();

	f = fopen (gift_conf_path ("shares"), "r");

	/* TODO - replace this mess. */
	if (!f)
		return;

	/* import line-by-line shares from disk */
	while (file_read_line (f, &buf))
	{
		char *pos = buf;
		char *md5, *size_str, *filename;
		char *filename_strip;

		/* TODO - use string_sep */
		trim_whitespace (pos);

		/* mtime */string_sep (&pos, " ");
		md5      = string_sep (&pos, " ");
		size_str = string_sep (&pos, " ");
		filename = pos;

		if (!filename)
			continue;

		size = strtoul (size_str, (char **)NULL, 10);

		filename_strip = html_strip (filename);
		file = share_new ("0.0.0.0", NODE (ft_self)->port,
		                  NODE (ft_self)->http_port, size, md5,
		                  filename_strip);
		free (filename_strip);

		share_local_add (file);
	}
}

/*****************************************************************************/

static int share_local_find_node (unsigned long key, FileShare *file,
								  char **filename)
{
	if (!filename || !*filename)
		return FALSE;

	if (!strcmp (*filename, file->file))
		*filename = NULL;

	return TRUE;
}

/* for security purposes, verify the supplied filename is being shared */
int share_local_find (char *filename)
{
	if (!filename)
		return FALSE;

	if (!strncmp (filename, DATA_DIR, strlen (DATA_DIR)))
		return TRUE;

	hash_table_foreach (ft_local_shares,
						(HashFunc) share_local_find_node, &filename);

	/* filename will be NULL if it was found */
	return (filename == NULL);
}

static void share_local_submit_write (Connection *c, FileShare *file,
									  void *udata)
{
	if (!file)
		return;

	ft_node_send (c, FT_SHARE_REQUEST, "%lu%s%s",
	              file->size, file->md5, file->file);
}

static void share_local_submit_destroy ()
{
}

void share_local_submit (Protocol *p, Connection *c)
{
	if (!ft_local_shares)
		return;

	/* TODO - hash_flatten is messy :/ */
	queue_add (c,
			   (QueueWriteFunc) share_local_submit_write,
			   (QueueWriteFunc) share_local_submit_destroy,
			   hash_flatten (ft_local_shares), NULL);
}

void share_local_foreach (HashFunc foreach_func, void *data)
{
	if (!ft_local_shares)
		return;

	hash_table_foreach (ft_local_shares, foreach_func, data);
}

/*****************************************************************************/

void share_stats_reset (Connection *c)
{
	NODE (c)->shares = 0;
	NODE (c)->megs   = 0;
}

static double share_stats_size (FileShare *file)
{
	unsigned long megs;

	megs = (((float)file->size / 1024.0) / 1024.0);

	return megs;
}

void share_stats_add (Connection *c, FileShare *file)
{
	NODE (c)->shares++;
	NODE (c)->megs += share_stats_size (file);
}

void share_stats_remove (Connection *c, FileShare *file)
{
	NODE (c)->shares--;
	NODE (c)->megs -= share_stats_size (file);
}

void share_stats_set (Connection *c, unsigned long users, unsigned long shares,
					  unsigned long megs)
{
	if (users)
		NODE (c)->users = users;

	if (shares)
		NODE (c)->shares = shares;

	if (megs)
		NODE (c)->megs = megs;
}
