/*
 * conf.c
 *
 * Copyright (C) 2001-2002 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <sys/stat.h>
#include <sys/types.h>

#include <fcntl.h>
#include <ctype.h>

#include "gift.h"

#include "conf.h"
#include "file.h"
#include "parse.h"

/*****************************************************************************/
/* helpers */

static char *config_parse_keypath (char *keypath, char **header_out,
								   char **key_out)
{
	char *header;
	char *key;
	char *value;
	char *defvalue;

	if ((defvalue = strchr (keypath, '=')))
		defvalue++;

	header = strdup (keypath);
	if (!(key = strchr (header, '/')))
	{
		free (header);
		return NULL;
	}

	*key++ = 0;

	if ((value = strchr (key, '=')))
		*value++ = 0;

	if (header_out)
		*header_out = header;

	if (key_out)
		*key_out = key;

	return defvalue;
}

/*****************************************************************************/

static ConfigHeader *config_header_new (char *line)
{
	ConfigHeader *confhdr;

	confhdr = malloc (sizeof (ConfigHeader));
	confhdr->name = strdup (line);
	confhdr->keys = NULL;

	return confhdr;
}

/*****************************************************************************/

static int cmp_config_header (ConfigHeader *a, char *b)
{
	if (!a->name || !b)
		return -1;

	return strcmp (a->name, b);
}

static char *config_lookup (Config *conf, ConfigHeader **confhdr_out,
                            char *header, char *key)
{
	ConfigHeader *confhdr;
	List *lheader;

	if (confhdr_out)
		*confhdr_out = NULL;

	lheader = list_find_custom (conf->headers, header,
	                            (CompareFunc) cmp_config_header);

	if (!lheader)
		return NULL;

	confhdr = lheader->data;

	if (confhdr_out)
		*confhdr_out = confhdr;

	return dataset_lookup (confhdr->keys, key);
}

/*****************************************************************************/
/* data insertion methods */

void config_set_str (Config *conf, char *keypath, char *value)
{
	char *header, *key, *defvalue;
	ConfigHeader *confhdr;

	defvalue = config_parse_keypath (keypath, &header, &key);

	/* find the header */
	config_lookup (conf, &confhdr, header, key);

	if (!confhdr)
	{
		conf->confhdr = config_header_new (header);
		conf->headers = list_append (conf->headers, conf->confhdr);

		confhdr = conf->confhdr;
	}

	dataset_insert (confhdr->keys, key, strdup (value));

	free (header);
}

void config_set_int (Config *conf, char *keypath, int value)
{
	char str[20];

	sprintf (str, "%i", value);

	config_set_str (conf, keypath, str);
}

/*****************************************************************************/
/* data retrieval methods */

char *config_get_str (Config *conf, char *keypath)
{
	char *header, *key, *defvalue;
	char *value;

	defvalue = config_parse_keypath (keypath, &header, &key);

	/* call the private locater function */
	value = config_lookup (conf, NULL, header, key);

	if (!value && defvalue)
	{
		value = STRDUP (defvalue);
		config_set_str (conf, keypath, value);
	}

	free (header);

	return value;
}

int config_get_int (Config *conf, char *keypath)
{
	char *value;

	value = config_get_str (conf, keypath);

	if (!value)
		return 0;

	return atoi (value);
}

/*****************************************************************************/

static void config_keys_read (Config *conf)
{
	char *line = NULL;
	char *key;
	char *value;

	while (file_read_line (conf->file, &line))
	{
		if (str_isempty (line) || *line == '[')
			return;

		key = line;
		if (!(value = strchr (key, '=')))
			continue;

		*value++ = 0;

		trim_whitespace (key);
		trim_whitespace (value);

		dataset_insert (conf->confhdr->keys, key, strdup (value));
	}
}

static void config_headers_read (Config *conf)
{
	char *line = NULL;

	conf->headers = NULL;

	if (!(conf->file = fopen (conf->path, "r")))
		return;

	while (file_read_line (conf->file, &line))
	{
		char *ptr;

		/* ignore blank lines */
		if (str_isempty (line) || *line == '#')
			continue;

		if (*line != '[')
			continue;

		ptr = strchr (line, ']');

		if (!ptr)
			continue;

		*ptr = 0;

		conf->confhdr = config_header_new (line + 1);
		conf->headers = list_append (conf->headers, conf->confhdr);

		config_keys_read (conf);
	}

	fclose (conf->file);
	conf->file = NULL;
}

Config *config_new (char *file)
{
	Config *conf;

	conf = malloc (sizeof (Config));

	conf->path = strdup (file);
	config_headers_read (conf);

	return conf;
}

static int free_key (unsigned long key, void *value, void *data)
{
	free (value);

	return 1;
}

void config_free (Config *conf)
{
	List *next;

	free (conf->path);

	if (conf->file)
		fclose (conf->file);

	while (conf->headers)
	{
		ConfigHeader *confhdr = conf->headers->data;

		next = conf->headers->next;

		hash_table_foreach_remove (confhdr->keys, (HashFunc) free_key, NULL);
		dataset_clear (confhdr->keys);
		free (confhdr->name);

		free (conf->headers);

		conf->headers = next;
	}
}

/*****************************************************************************/

Config *gift_config_new (char *module)
{
	char *file;

	if (!module)
		return NULL;

	if (!strcmp (module, "giFT"))
		file = gift_conf_path ("gift.conf");
	else
		file = gift_conf_path ("%s/%s.conf", module, module);

	return config_new (file);
}

/*****************************************************************************/

char *gift_conf_path(char *fmt, ...)
{
	static char retval[4096], tmp[4096];
	char *expanded;
	va_list args;

	va_start(args, fmt);
	vsnprintf (tmp, sizeof (retval) - 1, fmt, args);
	va_end(args);

	snprintf (retval, sizeof (retval) - 1, "~/%s/%s", GIFT_CONFIG_PATH, tmp);
	expanded = gift_expand_path (retval);

	gift_create_path (expanded);

	return expanded;
}

char *gift_find_home ()
{
	static char ph[4096];
	char *h;

	h = getenv ("HOME");
	if (h)
		sprintf (ph, "%s", h);
	else
	{
		h = getenv ("USER");

		if (h)
			sprintf (ph, "/home/%s", h);
		else
			sprintf (ph, ".");
	}

	return ph;
}

char *gift_expand_path (char *path)
{
	static char newp[4096];
	char *p1;

	path = (char *) strdup (path);

	/* TODO - convert to while loop */
	if ((p1 = strchr (path, '~')) != NULL)
	{
		char *p0 = path;

		*p1++ = 0;

		if (strlen(p0))
			sprintf (newp, "%s%s%s", p0, gift_find_home(), p1);
		else
			sprintf (newp, "%s%s", gift_find_home(), p1);
	}
	else
		strcpy (newp, path);	// possible buffer overflow

	free (path);

	return newp;
}

int gift_create_path (char *p)
{
	// create all subdirectories needed to open p
	char *p0, *p1;
	int ret = 0;

	p = (char *) strdup (p);

	p0 = p1 = p;

	while (*p1 == '/')
		p1++;

	while ((p1 = strchr (p1, '/')) != NULL)
	{
		*p1 = 0;

		/* Really should stat() here,
		 * but if the directory already exists,
		 * mkdir() will just fail anyway
		 */
		mkdir (p0, 0755);

		*p1 = '/';
		while (*p1 == '/')
			p1++;
	}

	free (p);

	return ret;
}
