/*
 * download.c
 *
 * Copyright (C) 2001-2002 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "gift.h"

#include "download.h"
#include "parse.h"
#include "event.h"
#include "conf.h"

/*****************************************************************************/

#define STATE_FILE_PREFIX "."
#define STATE_FILE_SUFFIX "state"

#define DIVIDE_THRESHOLD  500 /* bytes */

/*****************************************************************************/

static int download_report (Transfer *transfer)
{
	if (transfer->chunks)
	{
		if_event_reply (transfer->id, "transfer",
		                "id=i",       transfer->id,
		                "transmit=i", transfer->transmit,
		                "total=i",    transfer->total,
		                NULL);
	}

	return TRUE;
}

/*****************************************************************************/

static void download_handle_change (Transfer *transfer)
{
	if (!transfer->chunks)
	{
		GIFT_WARN (("no sources for %s", transfer->filename));
		download_report (transfer);
		download_free (transfer, TRUE);
		return;
	}
}

/*****************************************************************************/

static void download_event_close (IFEvent *event)
{
	if (event->c && event->c->data)
		if_event_reply (event->id, "transfer", "id=i", event->id, NULL);
}

Transfer *download_new (Connection *c, char *filename, unsigned long size)
{
	IFEventID id;
	Transfer *transfer;

	transfer = transfer_new (TRANSFER_DOWNLOAD, filename, size);

	/* every second dump stats */
	transfer->timer = timer_add (1, (TimerCallback) download_report,
	                             transfer);

	id = if_event_new (c, IFEVENT_TRANSFER, (IFEventFunc) download_event_close,
	                   transfer);

	transfer->id = id;

	return transfer;
}

/* TODO - this function is very incomplete */
void download_free (Transfer *transfer, int premature)
{
	List *ptr;

	trace ();

	for (ptr = transfer->chunks; ptr; ptr = list_next (ptr))
	{
		Chunk *chunk = ptr->data;

		if (premature)
		{
			/* notify the protocol that this chunk is finished */
			chunk_cancel (chunk);
		}
	}

	transfer_free (transfer);
}

/*****************************************************************************/

void download_stop (Transfer *transfer, int cancel)
{
	trace ();

	/* if cancel, remove state file */

	/* TODO - handle cancel */
	download_free (transfer, TRUE);
}

/*****************************************************************************/

void download_add_source (Transfer *transfer, char *location)
{
	List  *ptr;
	Chunk *chunk   = NULL;
	Chunk *largest = NULL;
	unsigned long start = 0;
	unsigned long stop  = 0;
	unsigned long diff  = 0;
	unsigned long largest_diff = 0;

	if (!location)
		return;

	for (ptr = transfer->chunks; ptr; ptr = ptr->next)
	{
		Chunk *tmp = ptr->data;

		/* partitioned chunk w/ a dead source */
		if (!tmp->source)
		{
			tmp->source = STRDUP (location);
			tmp->active = TRUE;

			/* indicate that we need a new chunk */
			chunk = tmp;

			break;
		}

		/* locate the largest available chunk */
		diff = tmp->stop - tmp->start - tmp->transmit;
		if (!largest || diff > largest_diff)
		{
			largest      = tmp;
			largest_diff = diff;
		}
	}

	/* do not redivide a chunk less than the desired threshold */
	if (largest && largest_diff < DIVIDE_THRESHOLD)
		return;

	if (largest)
	{
		diff = largest_diff / 2;
		stop = largest->stop;
		largest->stop -= diff;
		start = largest->stop;
	}
	else
	{
		stop = transfer->total;
	}

	/* a new division must be made */
	if (!chunk)
	{
		chunk = chunk_new (transfer, location, start, stop);
		chunk->p->download (chunk);
	}

	download_handle_change (transfer);
}

void download_remove_source (Transfer *transfer, char *location)
{
	List *ptr;

	if (!location)
		return;

	for (ptr = transfer->chunks; ptr; ptr = list_next (ptr))
	{
		Chunk *chunk = ptr->data;

		if (chunk->source && !strcmp (chunk->source, location))
		{
			if (chunk->start + chunk->transmit == chunk->stop)
			{
				transfer->chunks = list_remove (transfer->chunks, chunk);
				chunk_free (chunk);
			}
			else
			{
				printf ("dead segment from %s, %lu - %lu (%lu)\n",
				        chunk->source, chunk->start, chunk->stop, chunk->transmit);

				/* don't remove it from the list of chunks, it is a dead
				 * segment, simply nullify it */
				free (chunk->source);
				chunk->source = NULL;
				chunk->active = FALSE;
			}

			download_handle_change (transfer);

			return;
		}
	}

	printf ("failed to locate source %s!\n", location);
}

/*****************************************************************************/

/* writes the data to disk */
void download_write (Chunk *chunk, char *segment, size_t len)
{
	Transfer *transfer = chunk->transfer;
	unsigned long seek_pos  = chunk->start + chunk->transmit;
	signed long   remainder = chunk->stop - seek_pos;

	if (!chunk->active || len == 0)
		return;

	/* make sure we don't write too much */
	if (len > remainder)
		len = remainder;

	if (!transfer->f)
	{
		transfer->path = STRDUP (gift_conf_path ("incoming/%s",
		                                         transfer->filename));

		/* i'm not really sure why judge did this...if anyone has a problem
		 * with it, blame him */
		if (!(transfer->f = fopen (transfer->path, "r+")) &&
		    !(transfer->f = fopen (transfer->path, "w")))
		{
			download_free (transfer, TRUE);
			perror ("fopen");
			return;
		}

		GIFT_DEBUG (("opened %s", transfer->path));
	}

	/* incremental padding -- this is some ugly code here :) */
	if (seek_pos > transfer->max_seek)
	{
		int diff;

		/* seek to the current end */
		if (fseek (transfer->f, transfer->max_seek, SEEK_SET) == -1)
			perror ("fseek");

		/* pad the file with bs data */
		while ((diff = (seek_pos - transfer->max_seek)) > 0)
		{
			int n;

			if (len < diff)
				diff = len;

			n = fwrite (segment, sizeof (char), diff, transfer->f);
			if (n == -1)
			{
				perror ("fwrite");
				fclose (transfer->f);
				transfer->f = NULL;
				return;
			}

			transfer->max_seek += n;
		}
	}

	/* TODO - error handling */
	fseek (transfer->f, seek_pos, SEEK_SET);

	len = fwrite (segment, sizeof (char), MIN (len, remainder), transfer->f);

	chunk->transmit    += len; /* this chunk */
	transfer->transmit += len; /* total transfer stats */

	/* new seek max */
	if (seek_pos == transfer->max_seek)
		transfer->max_seek += len;

	remainder -= len;

	if (remainder <= 0)
	{
		chunk->active = FALSE;

		if (transfer->transmit >= transfer->total)
		{
			printf ("*** transfer completed successfully (%lu/%lu)\n",
			        transfer->transmit, transfer->total);
			download_report (transfer);
			download_free (transfer, FALSE);
			return;
		}

		download_add_source (transfer, chunk->source);

		printf ("*** chunk start (%lu) completed successfully (%lu/%lu)...%lu\n",
		        chunk->start, chunk->transmit, chunk->stop - chunk->start, remainder);
		download_handle_change (transfer);
	}
}
