/*
 * event.c
 *
 * Copyright (C) 2001-2002 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <time.h>
#include <sys/time.h>

#include <sys/types.h>

#include "gift.h"

#include "event.h"
#include "protocol.h"
#include "connection.h"

/*****************************************************************************/

static int loop_running = 1;

static List *inputs = NULL;
static List *timers = NULL;

/* it isnt safe to free input/timers at any time...so add them to this list to
 * be freed when it is safe */
static List *sources_free = NULL;

/*****************************************************************************/

void event_quit ()
{
	loop_running = 0;
}

/*****************************************************************************/

static time_t next_timer (time_t current_time)
{
	List *ptr;
	time_t lowest_time = 0;

	for (ptr = timers; ptr; ptr = ptr->next)
	{
		Timer *timer = ptr->data;

		if (ptr == timers || timer->expiration < lowest_time)
			lowest_time = timer->expiration;
	}

	if (lowest_time < current_time)
		return 0;

	return (lowest_time - current_time);
}

static void dispatch_timers (time_t current_time)
{
	List *safe_copy, *ptr;

	safe_copy = list_copy (timers);

	for (ptr = safe_copy; ptr; ptr = ptr->next)
	{
		Timer *timer = ptr->data;
		int ret;

		if (timer->expiration > current_time)
			continue;

		ret = (*timer->callback) (timer->data);

		if (!ret)
			timer_remove (timer->id);
		else
			timer->expiration = current_time + timer->interval;
	}

	list_free (safe_copy);
}

/*****************************************************************************/

static int prepare_fds (fd_set *rset, fd_set *wset, fd_set *xset)
{
	List *ptr;
	int maxfd = -1;

	FD_ZERO (rset);
	FD_ZERO (wset);
	FD_ZERO (xset);

	for (ptr = inputs; ptr; ptr = ptr->next)
	{
		Input *input = ptr->data;

		/* calculate maxfd for select () */
		if (input->fd > maxfd)
			maxfd = input->fd;

		/* fill fd sets */
		if (input->state & INPUT_READ)
			FD_SET (input->fd, rset);

		if (input->state & INPUT_WRITE)
			FD_SET (input->fd, wset);

		if (input->state & INPUT_EXCEPTION)
			FD_SET (input->fd, xset);
	}

	return maxfd;
}

static void dispatch_inputs (fd_set *rset, fd_set *wset, fd_set *xset)
{
	List *safe_copy, *ptr;

	/* these routines will fuck with the inputs list, so dont use it for
	 * processing */
	safe_copy = list_copy (inputs);

	for (ptr = safe_copy; ptr; ptr = ptr->next)
	{
		Input *input = ptr->data;
		InputCallback cb = input->callback;

		if (input->state & INPUT_READ      && FD_ISSET (input->fd, rset))
			(*cb) (input->protocol, input->data);

		if (input->state & INPUT_WRITE     && FD_ISSET (input->fd, wset))
			(*cb) (input->protocol, input->data);

		if (input->state & INPUT_EXCEPTION && FD_ISSET (input->fd, xset))
			(*cb) (input->protocol, input->data);
	}

	list_free (safe_copy);
}

/*****************************************************************************/

static void cleanup ()
{
	List *ptr;

	for (ptr = sources_free; ptr; ptr = ptr->next)
		free (ptr->data);

	list_free (sources_free);
	sources_free = NULL;
}

void event_loop ()
{
	fd_set rset, wset, xset;
	int maxfd;
	struct timeval next_event;
	time_t current_time;

	while (loop_running)
	{
		current_time = time (NULL);

		/* TODO - next_timer should probably have microsecond precision */
		next_event.tv_sec = next_timer (current_time);
		next_event.tv_usec = 0;
		maxfd = prepare_fds (&rset, &wset, &xset);

		if (select (maxfd + 1, &rset, &wset, &xset,
		            timers ? &next_event : NULL) < 0)
		{
			/* oh shit */
			perror ("select");
			continue;
		}

		/* dispatch */
		dispatch_inputs (&rset, &wset, &xset);
		dispatch_timers (current_time);

		cleanup ();
	}
}

/*****************************************************************************/
/* timer interface routines + helpers */

static Timer *timer_find (int id)
{
	List *ptr;

	for (ptr = timers; ptr; ptr = ptr->next)
	{
		Timer *timer = ptr->data;

		if (timer->id == id)
			return timer;
	}

	return NULL;
}

/* TODO - optimize */
static int timer_unique_id ()
{
	int id;

	for (id = 0; id < 65535; id++)
	{
		if (timer_find (id))
			continue;

		return id;
	}

	return -1;
}

int timer_add (time_t interval, TimerCallback callback, void *data)
{
	Timer *timer;
	time_t current_time;
	int uniq_id;

	if ((uniq_id = timer_unique_id ()) < 0)
		return uniq_id;

	timer = malloc (sizeof (Timer));

	timer->interval = interval;

	current_time      = time (NULL);
	timer->expiration = current_time + interval;

	timer->id = uniq_id;

	timer->callback = callback;
	timer->data     = data;

	timers = list_prepend (timers, timer);

	return uniq_id;
}

void timer_remove (int id)
{
	Timer *timer;

	if (id < 0)
		return;

	if (!(timer = timer_find (id)))
		return;

	/* TODO - double iteration */
	timers = list_remove (timers, timer);
	sources_free = list_prepend (sources_free, timer);
}

/*****************************************************************************/
/* input interface routines + helpers */

static Input *input_find (int fd, int condition, InputCallback cb)
{
	List *ptr;

	for (ptr = inputs; ptr; ptr = ptr->next)
	{
		Input *input = ptr->data;

		if (cb && input->callback != cb)
			continue;

		if (condition && input->state != condition)
			continue;

		if (input->fd == fd)
			return input;
	}

	return NULL;
}

void input_add (Protocol *p, Connection *c, InputState state,
                InputCallback callback)
{
	Input *input;

	if (!c || c->fd < 0)
		return;

	input = malloc (sizeof (Input));

	input->fd       = c->fd;
	input->state    = state;
	input->callback = callback;
	input->protocol = p;
	input->data     = c;

	inputs = list_prepend (inputs, input);
}

void input_remove_full (Connection *c, int condition, InputCallback cb)
{
	Input *input;

	if (!c || c->fd < 0)
		return;

	while ((input = input_find (c->fd, condition, cb)))
	{
		inputs = list_remove (inputs, input);
		sources_free = list_prepend (sources_free, input);
	}
}

void input_remove (Connection *c)
{
	input_remove_full (c, 0, NULL);
}
