/*
 * main.c
 *
 * TODO - better option handling
 *
 * Copyright (C) 2001-2002 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>
#include <errno.h>

#include "gift.h"
#include "event.h"

#include "conf.h"
#include "parse.h"
#include "protocol.h"

/*****************************************************************************/

#define INTERFACE_PORT 1213

static Dataset *options;   /* command-line parsed options */
List           *protocols; /* all fully loaded protocols  */

/*****************************************************************************/

static void usage (char *argv0, int exit_val)
{
	char *u = "Usage: %s [options] [protocol-lib...]\n"
              "\n"
              "    -h                  Help\n"
		      "    -v                  Version\n"
			  "    -d                  Fork to background\n"
              "<protocol-lib>          The file locations of the protocols to use\n";

	printf (u, argv0);
	exit (exit_val);
}

static void version ()
{
	/* simplified format (scriptability damnit) - jasta */
	printf ("%s-%s\n", PACKAGE, VERSION);
	exit (0);
}

/*****************************************************************************/

static int init ()
{
	struct stat st;
	char *gift_conf;

	options = NULL;

	/* COMMAND LINE OPTIONS */

	/* Write verbose */
	dataset_insert (options, "verbose", 0);

	/* Setup ~/.giFT, if not found */
	gift_conf = gift_expand_path ("~/.giFT/gift.conf");
	if (stat (gift_conf, &st))
	{
		FILE *f;

		gift_create_path (gift_conf);

		if (!(f = fopen (gift_conf, "w")))
		{
			perror ("fopen");
			GIFT_FATAL (("Failed to open %s for writing, please check permission\n",
						 gift_conf));
		}

		fprintf (f, "# gift.conf\n\n");
		fprintf (f, "[main]\n");
		fprintf (f, "plugins = %s/libOpenFT.so\n", PLUGIN_DIR);

		fclose (f);
	}

	return 1;
}

static List *parse (int argc, char **argv)
{
	List *plugin_fnames = NULL;

	/* use getopt to parse the command line */
	while (1)
	{
		int c;

		/* parse the command line */
		c = getopt (argc, argv, "hvVd");

		if (c == -1) break;

		switch (c)
		{
		 case 0:
			usage (argv[0], 1);
			break;
		 case 'h': /* help */
			usage (argv[0], 0);
			break;
		 case 'v': /* version */
		 case 'V':
			version (argv[0]);
			break;
		 case 'd': /* daemon */
			{
				pid_t cpid;

				if ((cpid = fork ()) != 0)
				{
					printf ("giFT: %u\n", cpid);
					_exit (0);
				}

				setsid ();

				close (STDIN_FILENO);
				close (STDOUT_FILENO);
				close (STDERR_FILENO);
			}
			break;
		 default:  { usage (argv[0], 1); } break;
		}
	}

	while (optind < argc)
		plugin_fnames = list_append (plugin_fnames, argv[optind++]);

	return plugin_fnames;
}

static List *pre_load (char *module, List *fnames)
{
	Config *conf;
	char *pstr, *p, *p0;

	if (!(conf = gift_config_new (module)))
		GIFT_ERROR (("failed to load configuration for module %s", module));

	p = config_get_str (conf, "main/plugins");

	if (p)
	{
		p0 = pstr = strdup (p);

		while ((p = string_sep (&pstr, ":")))
		{
			if (!p)
				continue;

			fnames = list_append (fnames, strdup (p));
		}

		free (p0);
	}

	config_free (conf);

	return fnames;
}

static int load_protocols (List *fnames)
{
	Protocol *p;
	List *temp;

	GIFT_DEBUG (("Loading protocols... "));

#if 0
	if (!(p = protocol_new ()))
		GIFT_FATAL (("Couldn't create new protocol, out of memory\n"));

	if (!(p = dc_init (p)))
		GIFT_FATAL (("Couldn't load client-daemon protocol\n"));

	protocols = list_append (protocols, p);
#endif

	/* load each plugin in the list */
	for (temp = fnames; temp; temp = temp->next)
	{
		if (!(p = protocol_load ((char*)(temp->data))))
			GIFT_ERROR (("\nFailed to load protocol: %s\n", (char *)temp->data));

		protocols = list_append (protocols, p);
	}

	if (!fnames)
		GIFT_WARN (("no protocols loaded"));

	return 1;
}

static int config (const char *module)
{
	Config *conf;

	GIFT_DEBUG (("Parsing configuration file... "));

	if (!(conf = gift_config_new ("giFT")))
		GIFT_ERROR (("failed to load main conf file"));

	config_free (conf);

	return 1;
}

/*****************************************************************************/

/* TODO: this needs to cleanup better */
static void shutdown ()
{
	List *temp;

	trace ();

	dataset_clear (options);

	for (temp = protocols; temp; temp = temp->next)
		protocol_unload (temp->data);

	protocols = list_free (protocols);

	exit (0);
}

/*****************************************************************************/
int main (int argc, char **argv)
{
	List *protocol_fnames = NULL;

	if (!init ())
		GIFT_FATAL (("Couldn't initialize the daemon\n"));

	signal (SIGPIPE, SIG_IGN);
	signal (SIGINT,  shutdown);

	/* atexit (shutdown); */

	protocol_fnames = parse (argc, argv);
	protocol_fnames = pre_load ("giFT", protocol_fnames);

	if (!protocol_fnames)
		usage (argv[0], 1);

	if (!(interface_init (INTERFACE_PORT)))
		GIFT_FATAL (("Couldn't load interface subsystem"));

	if (!(load_protocols (protocol_fnames)))
		GIFT_FATAL (("Couldn't load the protocols"));

	if (!config ("giFT"))
		GIFT_FATAL (("Couldn't parse the configuration file"));

	/* main event loop */
	event_loop ();

	shutdown ();

	return 0;
}
